/* =========================================================[ Preferences.cc ]=
 * Preferences.cc - set player preferences
 * Copyright (C) 2003 Hubert Chan <hubert@uhoreg.ca>
 *
 * ----------------------------------------------------------------( License )-
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * ------------------------------------------------------------( Description )-
 *
 * This file maintains a set of user preferences, and allows the user to change
 * them via an on-screen menu.  The preferences have a simple model: each
 * preference has a name, and a discrete set of possible values.  The values
 * are ordered (possibly in a circular order), and have string representations.
 *
 * ------------------------------------------------------------------------- */

#include <alsaplayer/prefs.h>

#include "Preferences.hh"

#include "debug.hh"

namespace {
  // a simple string-based prefs type.  Just a list of values.  The default
  // value is the first.  It is saved as lirc.<name>.
  class SimpleStringPref : public Preferences::Pref
  {
    const SessionData &sessiondata;
    std::vector<std::string> valueNames;
    int value;

    void (*callback)(const SessionData &, int);

  public:
    SimpleStringPref(const SessionData &_sd, const std::string &_name,
		     const std::vector<std::string> &_valueNames,
		     void (*_callback)(const SessionData &, int))
      : Preferences::Pref(_name),
	sessiondata(_sd),
	valueNames(_valueNames),
	value(0),
	callback(_callback)
      {
	// load the saved value
	std::string val = prefs_get_string(ap_prefs, "lirc",
					   const_cast<char*>(getName().c_str()),
					   const_cast<char*>(valueNames[0].c_str()));
	for (unsigned int i = 0; i < valueNames.size(); i++)
	{
	  if (val == valueNames[i])
	  {
	    value = i;
	    break;
	  }
	}
	callback(sessiondata, value);
      }

    const std::string &getValue() const { return valueNames[value]; }
    void nextValue()
      {
	value++; value %= valueNames.size();
	prefs_set_string(ap_prefs, "lirc",
			 const_cast<char*>(getName().c_str()),
			 const_cast<char*>(valueNames[value].c_str()));
	callback(sessiondata, value);
      }
    void prevValue()
      {
	value += valueNames.size()-1; value %= valueNames.size();
	prefs_set_string(ap_prefs, "lirc",
			 const_cast<char*>(getName().c_str()),
			 const_cast<char*>(valueNames[value].c_str()));
	callback(sessiondata, value);
      }
  };

  // a simple boolean prefs type.  True or false.  The default value is false,
  // unless otherwise specified.  It is saved as lirc.<name>.
  class SimpleBoolPref : public Preferences::Pref
  {
    const SessionData &sessiondata;
    bool value;

    void (*callback)(const SessionData &, bool);

  public:
    SimpleBoolPref(const SessionData &_sd, const std::string &_name,
		   void (*_callback)(const SessionData &, bool),
		   bool defVal = false)
      : Preferences::Pref(_name),
	sessiondata(_sd),
	callback(_callback)
      {
	// load the saved value
	value = prefs_get_bool(ap_prefs, "lirc",
			       const_cast<char*>(getName().c_str()), defVal);
	callback(sessiondata, value);
      }

    const std::string &getValue() const
      {
	static std::string yes("yes"), no("no");
	return value ? yes : no;
      }
    void nextValue()
      {
	value = !value;
	prefs_set_bool(ap_prefs, "lirc", const_cast<char*>(getName().c_str()),
		       value);
	callback(sessiondata, value);
      }
    void prevValue()
      {
	value = !value;
	prefs_set_bool(ap_prefs, "lirc", const_cast<char*>(getName().c_str()),
		       value);
	callback(sessiondata, value);
      }
  };

  void loopSongCallback(const SessionData &sessiondata, bool value)
  {
    if (value)
      sessiondata.playlist->LoopSong();
    else
      sessiondata.playlist->UnLoopSong();
  }

  void loopPlaylistCallback(const SessionData &sessiondata, bool value)
  {
    if (value)
      sessiondata.playlist->LoopPlaylist();
    else
      sessiondata.playlist->UnLoopPlaylist();
  }

  void autoAdvanceCallback(const SessionData &sessiondata, bool value)
  {
    if (value)
      sessiondata.playlist->UnPause();
    else
      sessiondata.playlist->Pause();
  }
}

Preferences::Preferences(const SessionData &sd)
{
  prefs.push_back(new SimpleBoolPref(sd, "loop_song", loopSongCallback));
  prefs.push_back(new SimpleBoolPref(sd, "loop_playlist", loopPlaylistCallback));
  prefs.push_back(new SimpleBoolPref(sd, "auto_advance", autoAdvanceCallback, true));
}

Preferences::~Preferences()
{
  for (unsigned int i = 0; i < prefs.size(); i++)
  {
    delete prefs[i];
  }
}
