/* ============================================================[ Selector.cc ]=
 * Selector.cc - select an item from a list
 * Copyright (C) 2002 Hubert Chan <hubert@uhoreg.ca>
 *
 * ----------------------------------------------------------------( License )-
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * ------------------------------------------------------------( Description )-
 *
 * This file maintains a list and allows the user to select an item.  It also
 * allows the user to go down into sublists.  This is used for inserting songs
 * into the playlist.
 *
 * ------------------------------------------------------------------------- */

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <cctype>
#include <algorithm>
#include <cstdlib>
#include <typeinfo>

#include <alsaplayer/prefs.h>

#include "Selector.hh"

// ---------------------------------------------------( Items and generators )-
class Selector::ListGenerator
{
public:
  virtual ~ListGenerator() { }

  virtual void populateList(std::vector<ListItem*> &selectionList) = 0;
};

namespace {
  // for sorting the selection list
  bool compareItems (const Selector::ListItem *a, const Selector::ListItem *b)
  { // case insensitive compare of the two names
    // mostly stolen from Stroustrup, _C++ Programming Language_, 3rd ed.,
    // p. 591
    std::string::const_iterator ai = a->getName().begin();
    std::string::const_iterator bi = b->getName().begin();

    while (ai != a->getName().end() && bi != b->getName().end())
    {
      if (toupper(*ai) != toupper(*bi)) return (toupper(*ai) < toupper(*bi));
      ++ai; ++bi;
    }

    return a->getName().length() < b->getName().length();
  }

  std::string filenameToName(const std::string &filename)
  {
    // find the last slash (i.e. the beginning of the file name)
    unsigned int slashpos = filename.rfind('/');
    if (slashpos == std::string::npos)
      slashpos = 0;
    else
      slashpos++;

    // find the first dot in the file name (i.e. the beginning of the
    // extension)
    unsigned int dotpos = filename.find('.', slashpos);
    if (dotpos == std::string::npos)
      dotpos = filename.length();

    std::string name;

    for (unsigned int i = slashpos; i < dotpos; i++)
    {
      switch (filename[i])
      {
      case '_':
	name.push_back(' ');
	break;
      case '%':
      {
	// interpret %.. as a hex-encoded character
	char c = '\0';
	if (++i >= filename.length() || !std::isxdigit(filename[i]))
	{ // just a lone % or not a valid hex number
	  name.push_back('%');
	  i--;
	  break;
	}
	if (filename[i] >= '0' && filename[i] <= '9')
	  c = filename[i] - '0';
	else if (filename[i] >= 'a' && filename[i] <= 'f')
	  c = filename[i] - 'a' + 0xa;
	else
	  c = filename[i] - 'A' + 0xa;
	if (++i >= filename.length() || !std::isxdigit(filename[i]))
	{ // %. -- don't interpret this as a hex number
	  name.push_back('%');
	  i--; i--;
	  break;
	}
	c *= 0x10;
	if (filename[i] >= '0' && filename[i] <= '9')
	  c |= filename[i] - '0';
	else if (filename[i] >= 'a' && filename[i] <= 'f')
	  c |= filename[i] - 'a' + 0xa;
	else
	  c |= filename[i] - 'A' + 0xa;
	name.push_back(c);
	break;
      }
      default:
	name.push_back(filename[i]);
      }
    }

    return name;
  }


  // ................................................../ Files & Directories \.
  // a set of song files organized in a directory (possibly nested)
  class SongFileItem : public Selector::ListItem
  {
    std::string filename;

  public:
    SongFileItem(const std::string &_filename)
      : Selector::ListItem(filenameToName(_filename)),
	filename(_filename)
      { }

    void addToPlaylist(Playlist *playlist, int pos) const
      {
	playlist->Insert(filename, pos);
      }
  };

  class SongDirectoryItem : public Selector::ListItem
  {
    std::string directory;
  public:
    SongDirectoryItem(const std::string &_directory)
      : Selector::ListItem(filenameToName(_directory) + '/'),
	directory(_directory)
      { }

    Selector::ListGenerator *sublistGenerator() const;
  };

  bool dirCompareItems (const Selector::ListItem *a, const Selector::ListItem *b)
  { // case insensitive compare of the two names, directories go first
    // Ick!  Uses RTTI!
    if (typeid(*a) == typeid(SongDirectoryItem)
	&& typeid(*b) != typeid(SongDirectoryItem))
      return true;
    if (typeid(*b) == typeid(SongDirectoryItem)
	&& typeid(*a) != typeid(SongDirectoryItem))
      return false;

    return compareItems(a,b);
  }

  class SongDirectoryGenerator : public Selector::ListGenerator
  {
    std::string directory;
  public:
    SongDirectoryGenerator(const std::string &_directory)
      : directory(_directory)
      { }

    void populateList(std::vector<Selector::ListItem*> &selectionList)
      {
	DIR *dp = opendir(directory.c_str());
	if (!dp)
	  return;
	struct dirent *entry;

	while ((entry = readdir(dp)))
	{
	  std::string filename = directory + "/" + entry->d_name;
	  if (filename.length() >= 4 &&
	      filename.substr(filename.length() - 4) == ".m3u")
	    continue; // skip .m3u files
	  if (entry->d_name[0] == '.')
	    continue; // skip hidden files
	  if (filename[filename.length()-1] == '~')
	    continue; // skip backup files

	  struct stat statbuf;
	  if (stat(filename.c_str(), &statbuf) < 0)
	    continue; // could not stat file -- ignore

	  if (S_ISREG(statbuf.st_mode))
	    selectionList.push_back(new SongFileItem(filename));
	  else if (S_ISDIR(statbuf.st_mode))
	    selectionList.push_back(new SongDirectoryItem(filename));
	}

	closedir(dp);
	sort(selectionList.begin(), selectionList.end(), dirCompareItems);
      }
  };


  Selector::ListGenerator *SongDirectoryItem::sublistGenerator() const
  {
    return new SongDirectoryGenerator(directory);
  }


  // ............................................................/ Playlists \.
  class M3UPlaylistItem : public Selector::ListItem
  {
    std::string filename;

  public:
    M3UPlaylistItem(const std::string &_filename)
      : Selector::ListItem(filenameToName(_filename)),
	filename(_filename)
      { }

    void addToPlaylist(Playlist *playlist, int pos) const
      {
	playlist->Load(filename, pos, false);
      }
  };

  class PlaylistsGenerator : public Selector::ListGenerator
  {
  public:
    void populateList(std::vector<Selector::ListItem*> &selectionList)
      {
	std::string home = std::getenv("HOME");
	std::string playlist_paths = prefs_get_string(ap_prefs, "lirc",
						      "playlist_path",
						      "$HOME/music");
	std::vector<std::string> paths;

	std::string::size_type i = 0, last_i = 0;
	do
	{
	  last_i = i;
	  i = playlist_paths.find(':', last_i);
	  std::string path = playlist_paths.substr(last_i, i - last_i);
	  if (path.substr(0, 5) == "$HOME")
	    path.replace(0, 5, home);
	  paths.push_back(path);
	} while (i != std::string::npos && i++);

	for (i = 0; i < paths.size(); i++)
	{
	  DIR *dp = opendir(paths[i].c_str());
	  if (!dp)
	    continue;
	  struct dirent *entry;

	  while ((entry = readdir(dp)))
	  {
	    std::string filename = paths[i] + "/" + entry->d_name;
	    // I bet this conditional is really nasty performance-wise
	    if (filename.length() < 5 ||
		filename.substr(filename.length() - 4) != ".m3u")
	      continue; // only consider .m3u files

	    struct stat statbuf;
	    if (stat(filename.c_str(), &statbuf) < 0)
	      continue; // could not stat file -- ignore

	    if (!S_ISREG(statbuf.st_mode))
	      continue; // not a regular file

	    selectionList.push_back(new M3UPlaylistItem(filename));
	  }
	  closedir(dp);
	}

	sort(selectionList.begin(), selectionList.end(), compareItems);
      }
  };


  // ......................................................./ Top-level menu \.
  class MusicDirectoryItem : public Selector::ListItem
  {
  public:
    MusicDirectoryItem() : ListItem("Music") { }

    Selector::ListGenerator *sublistGenerator() const
      {
	std::string home = std::getenv("HOME");
	std::string music_path = prefs_get_string(ap_prefs, "lirc",
						  "music_path",
						  "$HOME/music");
	if (music_path.substr(0, 5) == "$HOME")
	  music_path.replace(0, 5, home);
	return new SongDirectoryGenerator(music_path);
      }
  };

  class PlaylistsItem : public Selector::ListItem
  {
  public:
    PlaylistsItem() : ListItem("Playlists") { }

    Selector::ListGenerator *sublistGenerator() const
      { return new PlaylistsGenerator(); }
  };

  class BaseGenerator : public Selector::ListGenerator
  {
  public:
    void populateList(std::vector<Selector::ListItem*> &selectionList)
      {
	selectionList.push_back(new MusicDirectoryItem());
	selectionList.push_back(new PlaylistsItem());
	//selectionList.push_back(new Selector::ListItem("CDs (not implemented)"));
      }
  };
}

Selector::Selector()
{
  history.push(new BaseGenerator());
  history.top()->populateList(selectionList);
}

Selector::~Selector()
{
  while (!history.empty())
  {
    delete history.top();
    history.pop();
  }
  clear();
}

void Selector::clear()
{
  for (std::vector<ListItem*>::iterator i = selectionList.begin();
       i != selectionList.end(); ++i)
  {
    delete *i;
  }
  selectionList.clear();
}

bool Selector::loadSublist(int selection)
{
  ListGenerator *generator = selectionList[selection]->sublistGenerator();

  if (generator)
  {
    history.push(generator);

    clear();

    generator->populateList(selectionList);

    return true;
  }
  else
    return false;
}

bool Selector::backout()
{
  // make sure we don't back out too far
  if (history.size() > 1)
  {
    delete history.top();
    history.pop();

    clear();

    history.top()->populateList(selectionList);

    return true;
  }
  else
    return false;
}

void Selector::addToPlaylist(int selection, Playlist *playlist, int pos)
{
  return selectionList[selection]->addToPlaylist(playlist, pos);
}
