/* =========================================================[ TextDisplay.hh ]=
 * TextDisplay.hh - a text-based display
 * Copyright (C) 2002 Hubert Chan <hubert@uhoreg.ca>
 *
 * ----------------------------------------------------------------( License )-
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * ------------------------------------------------------------( Description )-
 *
 * This file implements a text-based display, such as XOSD, an external
 * character LCD or VFD display, or an ncurses display.
 *
 * ------------------------------------------------------------------------- */

// enable the UNIX98 standard, so that we can set the mutex type
#define _XOPEN_SOURCE 500

#include <sstream>
#include <iomanip>
//#include <ios>
#include <unistd.h>
#include <sys/time.h>

#include <alsaplayer/prefs.h>

#include "TextDisplay.hh"

extern "C" {
#include <xosd.h>
}

#include "debug.hh"


// ------------------------------------------------------( Utility functions )-
template <class T>
inline std::string slider_string(T min, T max, T value, int size)
{
  std::string str(size, '-');
  int pos = int((value - min) * size / (max - min));
  if (pos < 0)
    pos = 0;
  else if (pos >= size)
    pos = size-1;
  str[pos] = '|';
  return str;
}

inline int timeval_gt (struct timeval *t1, struct timeval *t2)
{
  if (t1->tv_sec < t2->tv_sec)
    return 0;
  if (t1->tv_sec == t2->tv_sec && t1->tv_usec <= t2->tv_usec)
    return 0;
  return 1;
}

inline std::string make_string(const PlayItem &item)
{
  if (const_cast<PlayItem&>(item).Parsed())
  {
    return std::string(item.title) + " - " + item.artist;
  }
  else
  {
    return item.filename;
  }
}

namespace {
  static std::string playlistModeStrings[] = {"Play Song",
					      "Insert",
					      "Clear Playlist",
					      "Delete Song"};


  // ------------------------------------------------------( Display drivers )-
  class XOSD_Display : public TextDisplay::DisplayDriver
  {
    xosd *osd;
  public:
    XOSD_Display()
      : DisplayDriver(prefs_get_int(ap_prefs, "lirc", "xosd_width", 40),
                      prefs_get_int(ap_prefs, "lirc", "xosd_height", 4))
      {
	osd = xosd_create(getHeight());

        xosd_set_font(osd,
		      prefs_get_string(ap_prefs, "lirc", "xosd_font",
				       "fixed"));
        xosd_set_colour(osd,
			prefs_get_string(ap_prefs, "lirc", "xosd_colour",
					 "Red"));
        char *position = prefs_get_string(ap_prefs, "lirc", "xosd_pos", "bottom"
);
        xosd_pos pos = strcmp(position, "top") == 0 ? XOSD_top
	  : strcmp(position, "middle") == 0 ? XOSD_middle : XOSD_bottom;
	xosd_set_pos(osd, pos);

	xosd_set_timeout(osd, -1);
      }

    ~XOSD_Display()
      {
        xosd_destroy(osd);
      }

    void print(unsigned int line, const std::string &str)
      {
        xosd_display(osd, line, XOSD_string, str.c_str());
      }

    void hide()
      {
        xosd_hide(osd);
      }

    void show()
      {
        xosd_show(osd);
      }
  };

  TextDisplay::DisplayDriver * create_driver()
  {
    std::string display = prefs_get_string(ap_prefs, "lirc", "text_display_type", "xosd");
    return new XOSD_Display();
  }
}


// ----------------------------------------------------------------( Windows )-
namespace {
  std::list<TextDisplay::Window*> dummy_list;
}

class TextDisplay::Window
{
  TextDisplay &parent;

protected:
  pthread_mutex_t window_lock;
  std::vector<std::string> screenLines;
  std::list<Window*>::iterator activePosition;

public:
  Window(TextDisplay &parent);
  virtual ~Window();

  void print(unsigned int line, const std::string &str);
  void hide();
  virtual void show();

  bool isActive() const { return activePosition != dummy_list.begin(); }
  bool isVisible() const { return getActiveWindows().back() == this; }

protected:
  virtual void coveredCallback() { }

  // allow subclasses to access stuff from the parent
  void lock_display() const { pthread_mutex_lock(&(parent.display_lock)); }
  void unlock_display() const { pthread_mutex_unlock(&(parent.display_lock)); }
  void lock_window() { pthread_mutex_lock(&window_lock); }
  void unlock_window() { pthread_mutex_unlock(&window_lock); }

  std::list<Window*> &getActiveWindows() const { return parent.activeWindows; }
  DisplayDriver *getDriver() const { return parent.driver; }
  SessionData &getSessionData() const { return parent.sessiondata; }
};

TextDisplay::Window::Window(TextDisplay &_parent)
  : parent(_parent),
    screenLines(parent.getHeight()),
    activePosition(dummy_list.begin())
{
  pthread_mutexattr_t attr;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init(&window_lock, &attr);
  pthread_mutexattr_destroy(&attr);
}

TextDisplay::Window::~Window()
{
  hide();

  pthread_mutex_destroy(&window_lock);
}

void TextDisplay::Window::print(unsigned int i, const std::string &str)
{
  lock_window();
  if (i < unsigned(getDriver()->getHeight()))
    screenLines[i] = str;

  lock_display();
  if (isVisible())
    getDriver()->print(i, str);

  unlock_display();
  unlock_window();
}

void TextDisplay::Window::hide()
{
  lock_window();
  lock_display();
  if (isActive())
  {
    getActiveWindows().erase(activePosition);
    activePosition = dummy_list.begin();
  }

  if (getActiveWindows().size() == 0)
  { // no active windows -- hide the display
    getDriver()->hide();
  }
  else
  { // display the previous window
    for (unsigned int i = 0; i < screenLines.size(); i++)
      getDriver()->print(i, getActiveWindows().back()->screenLines[i]);
  }
  unlock_display();
  unlock_window();
}

void TextDisplay::Window::show()
{
  lock_window();
  if (isVisible())
  {
    unlock_window();
    return; // already on top -- do nothing
  }
  lock_display();
  if (isActive())
    // already active, but not visible -- remove its current entry
    getActiveWindows().erase(activePosition);

  // if list is not empty, call the "covered" callback of the top window
  if (getActiveWindows().size() != 0)
  {
   getActiveWindows().back()->coveredCallback();
  }

  // add to the top of the list
  activePosition = getActiveWindows().insert(getActiveWindows().end(), this);

  // display
  for (unsigned int i = 0; i < screenLines.size(); i++)
    getDriver()->print(i, screenLines[i]);
  getDriver()->show();
  unlock_display();
  unlock_window();
}


// -----------------------------------------------------------( Window types )-
// .........................................................../ Timed Window \.
class TimedWindow : public TextDisplay::Window
{ // a window which automatically disappears after a set time
  pthread_t timer_thread;
  pthread_cond_t timer_kill;

  int timeout;
  struct timeval hide_time;

  static void *timer(TimedWindow *);

public:
  TimedWindow(TextDisplay &_parent, int _timeout)
    : Window(_parent), timeout(_timeout)
    {
      hide_time.tv_sec = 0;
      pthread_cond_init(&timer_kill, NULL);
    }

  ~TimedWindow()
    {
      DEBUG2 << "Destroying a TimedWindow\n";
      lock_window();
      if (hide_time.tv_sec)
      {
	DEBUG2 << "Killing the timer thread\n";
	hide_time.tv_sec = 0;
	pthread_cond_signal(&timer_kill);
	unlock_window();
	pthread_join(timer_thread, NULL);
      }
      else
	unlock_window();
      pthread_cond_destroy(&timer_kill);
      DEBUG2 << "Done destroy TimedWindow\n";
    }

  void startTimer();

  void show()
    {
      lock_window();
      Window::show();
      startTimer();
      unlock_window();
    }
};

void *TimedWindow::timer(TimedWindow *window)
{
  window->lock_window();

  struct timeval curr_time;
  struct timespec timeout;

  while ( (gettimeofday(&curr_time, NULL),
	   timeval_gt(&(window->hide_time), &curr_time)) )
  {
    timeout.tv_sec = window->hide_time.tv_sec;
    timeout.tv_nsec = window->hide_time.tv_usec * 1000L;
    pthread_cond_timedwait(&(window->timer_kill),
                           &(window->window_lock),
                           &timeout);
  }

  window->hide_time.tv_sec = 0;

  // restore display
  window->hide();

  window->unlock_window();

  pthread_exit(NULL);
}

void TimedWindow::startTimer()
{
  lock_window();
  if (timeout)
  {
    if (hide_time.tv_sec == 0) // no thread currently active
      pthread_create(&timer_thread, NULL, (void*(*)(void*))TimedWindow::timer,
		     this);
    gettimeofday(&hide_time, NULL);
    hide_time.tv_sec += timeout;
  }
  unlock_window();
}


// ........................................................../ Player Window \.
class TextDisplay::PlayerWindow : public TimedWindow
{
public:
  PlayerWindow(TextDisplay &parent)
    : TimedWindow(parent, prefs_get_int(ap_prefs, "lirc", "display_timeout", 3))
    { }

  void show()
    { // put the player at the bottom of the stack when it's shown, instead
      // of on top
      lock_window();
      lock_display();
      if (isActive())
      { // already active -- just restart the timer
	unlock_display();
	startTimer();
	unlock_window();
	return;
      }

      activePosition = getActiveWindows().insert(getActiveWindows().begin(),
						 this);

      if (isVisible())
      {
	for (unsigned int i = 0; i < screenLines.size(); i++)
	  getDriver()->print(i, screenLines[i]);
	getDriver()->show();
      }

      unlock_display();
      startTimer();
      unlock_window();
    }

  enum status_t { STOPPED, PLAYING, PAUSED };

public:
  void setStatus(status_t status)
    {
      static std::string statusStrings[] = { "stopped", "playing", "paused" };
      print(0, statusStrings[status]);
      show();
    }

  void setCurrentSong(const PlayItem &item)
    {
      if (&item == NULL)
      {
	for (int i = 1; i < getDriver()->getHeight(); i++)
	  print(i, "");
	return;
      }

      if (const_cast<PlayItem&>(item).Parsed())
      {
	if (getDriver()->getHeight() <= 3)
	  print(1, item.title + " - " + item.artist);
	else
	{
	  print(1, item.title);
	  print(2, item.artist);
	  if (getDriver()->getHeight() > 4)
	    print(3, item.album);
	}
      }
      else
      {
	print(1, item.filename);
	for (int i = 2; i < getDriver()->getHeight(); i++)
	{
	  print(i, "");
	}
      }

      show();
    }

  void positionNotify(int frame)
    {
      if (getDriver()->getHeight() < 3)
	return; // screen too small to show position data
      int num_frames = getSessionData().coreplayer->GetFrames();
      if (num_frames <= 0)
	return;
      int current_time = getSessionData().coreplayer->GetCurrentTime() / 100;

      std::ostringstream str;
      str << '['
	  << slider_string(0, num_frames, frame, getDriver()->getWidth() - 8)
	  << "] " << std::setw(2) << (current_time/60) << ':'
	  << std::setfill('0') << std::setw(2) << (current_time%60);
      print(getDriver()->getHeight()-1, str.str());
    }
};


// .........................................................../ Popup Window \.
class TextDisplay::PopupWindow : public TimedWindow
{
public:
  PopupWindow(TextDisplay &parent)
    : TimedWindow(parent, prefs_get_int(ap_prefs, "lirc", "popup_timeout", 1))
    { }

  void coveredCallback() { hide(); }
};


// ............................................................/ List Window \.
class ListWindow : public TextDisplay::Window
{
  int displayTop;
  int y;

public:
  ListWindow(TextDisplay &parent, int _y = 0)
    : Window(parent),
      displayTop(0),
      y(_y)
    {}

protected:
  // which line is selected?
  virtual int getSelection() = 0;
  // how big is the list?
  virtual int getSize() = 0;
  // how do you display the current line?
  virtual std::string getLine(int) = 0;
  // how high should the display be?
  virtual int getHeight() = 0;

private:
  // returns whether or not the screen has changed
  bool updateDisplayTop()
    {
      int selected = getSelection();
      int height = getHeight();
      if (selected < displayTop)
	// selection comes before current display -- put at the bottom
	displayTop = std::max(0, selected - height + 1);
      else if (selected >= displayTop + height)
	// selection comes after current display -- put at the top
	displayTop = selected;
      else
	return false;
      return true;
    }

public:
  void update(int top = 0, int bottom = 65536)
    {
      DEBUG3 << "ListWindow::update(" << top << ',' << bottom << ")\n";
      lock_window();
      int height = getHeight();
      if (getSize() <= 0)
      {
	DEBUG2 << "List is empty\n";
	print(y, "<empty>");
	for (int i = 1; i < height; i++)
	  print(i+y, "");
	unlock_window();
	DEBUG3 << "exiting ListWindow::update\n";
	return;
      }

      bottom = std::min(bottom - displayTop + 1, height);
      top = std::max(top - displayTop, 0);
      DEBUG3 << "Updating from" << top << " to " << bottom << '\n';
      for (int i = top + displayTop;
	   i < getSize() && top < bottom;
	   i++, top++)
      {
	DEBUG3 << ' ' << i << '\n';
	print(top+y, getLine(i));
      }

      // fill the rest with blank lines
      for (; top < bottom; top++)
	print(top+y, "");
      unlock_window();
      DEBUG3 << "exiting ListWindow::update\n";
    }

  void selectionChanged(int oldSelection)
    {
      lock_window();
      if (updateDisplayTop())
	update();
      else
      {
	update(oldSelection,oldSelection);
	int selection = getSelection();
	update(selection,selection);
      }
      unlock_window();
    }
};

// ......................................................../ Playlist Window \.
class TextDisplay::PlaylistWindow : public ListWindow
{
  int current, selected;
public:
  // if the screen is big enough, the last line will be reserved for showing
  // the mode
  bool bigScreen;

  PlaylistWindow(TextDisplay &parent)
    : ListWindow(parent),
      current(-1),
      selected(-1),
      bigScreen(getDriver()->getHeight() >= 3)
    { }

private:
  int getSelection()
    {
      return (selected < 0) ? current : selected;
    }

  int getSize()
    {
      if (getSessionData().playlistInterface->getSize())
	return getSessionData().playlistInterface->getSize() + 1;
      else
	return 0;
    }

  std::string getLine(int line)
    {
      if (line >= getSessionData().playlistInterface->getSize())
	return ' ' + std::string(1, (line == selected) ? '>' : ' ');
      else
	return std::string(1, (line == current) ? 'P' : ' ') +
	  std::string(1, (line == selected) ? '>' : ' ') +
	  make_string(getSessionData().playlistInterface->getItem(line));
    }

  int getHeight()
    {
      return getDriver()->getHeight()
	- (bigScreen ? 1 : 0);
    }

public:
  void setCurrentSong(int cur)
    {
      lock_window();
      int oldCurrent = current;
      current = cur;
      if (selected < 0)
	selectionChanged(oldCurrent);
      else
      {
	update(oldCurrent,oldCurrent);
	update(current,current);
      }
      unlock_window();
    }

  void setSelectedSong(int sel)
    {
      lock_window();
      int oldSelected = selected;
      selected = sel;
      selectionChanged(oldSelected);
      unlock_window();
    }
};


// ......................................................../ Selector Window \.
class TextDisplay::SelectorWindow : public ListWindow
{
  int selected;

public:
  SelectorWindow(TextDisplay &parent)
    : ListWindow(parent),
      selected(0)
    { }

private:
  int getSelection()
    {
      return selected;
    }

  int getSize()
    {
      return getSessionData().selector->getSize();
    }

  std::string getLine(int line)
    {
      return std::string(1, (int(line) == selected) ? '>' : ' ') +
	getSessionData().selector->getItemName(line);
    }

  int getHeight()
    {
      return getDriver()->getHeight();
    }

public:
  void setSelection(int sel)
    {
      lock_window();
      int oldSelected = selected;
      selected = sel;
      selectionChanged(oldSelected);
      unlock_window();
    }
};


// ....................................................../ Preference Window \.
class TextDisplay::PreferenceWindow : public ListWindow
{
  int selected;

public:
  PreferenceWindow(TextDisplay &parent)
    : ListWindow(parent),
      selected(0)
    { }

private:
  int getSelection()
    {
      return selected;
    }

  int getSize()
    {
      return getSessionData().preferences->getSize();
    }

  std::string getLine(int line)
    {
      return std::string(1, (int(line) == selected) ? '>' : ' ') +
	getSessionData().preferences->getItem(line).getName() + ": " +
	getSessionData().preferences->getItem(line).getValue();
    }

  int getHeight()
    {
      return getDriver()->getHeight();
    }

public:
  void setSelection(int sel)
    {
      lock_window();
      int oldSelected = selected;
      selected = sel;
      selectionChanged(oldSelected);
      unlock_window();
    }
};


// -----------------------------------------------------------( Text Display )-
TextDisplay::TextDisplay(SessionData &_sd)
  : Display(_sd),
    driver(create_driver()),
    playerWindow(new PlayerWindow(*this)),
    popupWindow(new PopupWindow(*this)),
    playlistWindow(new PlaylistWindow(*this)),
    selectorWindow(new SelectorWindow(*this)),
    preferenceWindow(new PreferenceWindow(*this))
{
  notifier_init();

  pthread_mutexattr_t attr;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init(&display_lock, &attr);
  pthread_mutexattr_destroy(&attr);

  if (sessiondata.coreplayer->IsPlaying())
    playerStarted();
  else
    playerStopped();
  setPlaylistMode(PLMODE_PLAY);
}

TextDisplay::~TextDisplay()
{
  delete playerWindow;
  delete popupWindow;
  delete playlistWindow;
  delete selectorWindow;
  delete preferenceWindow;

  delete driver;

  pthread_mutex_destroy(&display_lock);
}


// -----------------------------------------------( Core player notification )-
void TextDisplay::volumeChanged(int volume)
{
  std::ostringstream str;
  str << "vol: ["
      << slider_string(0, 100, volume, getWidth() - 11)
      << "] " << std::setw(3) << volume;
  popupWindow->print(getHeight()/2, str.str());
  popupWindow->show();
}

void TextDisplay::speedChanged(float speed)
{
  std::ostringstream str;
  str.setf(std::ios::fixed, std::ios::floatfield); // FIXME: make this standards compliant
  str << "speed: ["
      << slider_string(-2.0, 2.0, double(speed), getWidth() - 15)
      << "] " << std::setprecision(2) << std::setw(5) << speed;
  popupWindow->print(getHeight()/2, str.str());
  popupWindow->show();
}

void TextDisplay::panChanged(int pan)
{
  std::ostringstream str;
  str << "pan: ["
      << slider_string(-100, 100, pan, getWidth() - 11)
      << "] " << std::setw(4) << pan;
  popupWindow->print(getHeight()/2, str.str());
  popupWindow->show();
}

void TextDisplay::positionNotify(int frame)
{
  playerWindow->positionNotify(frame);
}

void TextDisplay::playerStarted()
{
  playerWindow->setStatus(PlayerWindow::PLAYING);
}

void TextDisplay::playerStopped()
{
  playerWindow->setStatus(PlayerWindow::STOPPED);
}


// --------------------------------------------------------( Player controls )-
void TextDisplay::playerPaused()
{
  playerWindow->setStatus(PlayerWindow::PAUSED);
}

void TextDisplay::playerUnpaused()
{
  playerWindow->setStatus(PlayerWindow::PLAYING);
}


// ----------------------------------------------------( Playlist management )-
void TextDisplay::setCurrentSong(int pos)
{
  if (pos < 0)
    playerWindow->setCurrentSong(*(PlayItem*)(NULL));
  else
    playerWindow->setCurrentSong(sessiondata.playlistInterface->getItem(pos));

  playlistWindow->setCurrentSong(pos);
}

void TextDisplay::setSelectedSong(int pos)
{
  playlistWindow->setSelectedSong(pos);
}

void TextDisplay::songsInserted(int pos)
{
  playlistWindow->update(pos);
}

void TextDisplay::songUpdated(int pos)
{
  playlistWindow->update(pos,pos);
}

void TextDisplay::songsRemoved(int pos)
{
  playlistWindow->update(pos);
}

void TextDisplay::playlistCleared()
{
  playlistWindow->update();
}

void TextDisplay::setPlaylistMode(playlistmode_t mode)
{
  if (playlistWindow->bigScreen)
  {
    int padding = getWidth() - 2 - playlistModeStrings[mode].length();
    std::string str = '<' + std::string(padding/2, ' ')
      + playlistModeStrings[mode] + std::string((padding+1)/2, ' ') + '>';
    playlistWindow->print(getHeight()-1, str);
  }
  else if (sessiondata.getMode() == MODE_PLAYLIST)
  {
    popupWindow->print(getHeight()/2, playlistModeStrings[mode]);
    popupWindow->show();
  }
}

void TextDisplay::setSelection(int pos)
{
  selectorWindow->setSelection(pos);
}

void TextDisplay::updateSelectionList()
{
  selectorWindow->update();
}


// ------------------------------------------------------- ( Preference menu )-
void TextDisplay::setPreference(int pos)
{
  preferenceWindow->setSelection(pos);
}

void TextDisplay::updatePreference(int pos)
{
  preferenceWindow->update(pos,pos);
}


// ------------------------------------------------------------( Mode change )-
void TextDisplay::modeChangedCallback()
{
  switch (sessiondata.getMode())
  {
  case MODE_PLAYER:
    playlistWindow->hide();
    selectorWindow->hide();
    preferenceWindow->hide();
    break;
  case MODE_PLAYLIST:
    playlistWindow->show();
    selectorWindow->hide();
    break;
  case MODE_SELECT:
    selectorWindow->update();
    selectorWindow->show();
    break;
  case MODE_MENU:
    preferenceWindow->update();
    preferenceWindow->show();
  default:
    break;
  }
}
